"""
    (c) 2020 Rodney Maniego Jr.
    Maguro
"""

import shlex
import threading

class Maguro(list):
    """ Parse DSV files into a 2D Python array. """
    def __init__(self, filepath=None, delimiter=",", newline=None, encoding="utf-8", autosave=True, quote_strings=False, allow_boolean=False, NaN="NaN"):
        self._lock = threading.RLock()
        with self._lock:
            self._filepath = filepath
            self._delimiter = delimiter
            self._newline = newline
            self._autosave = autosave
            self._quote_strings = quote_strings
            self._encoding = encoding
            self._allow_boolean = allow_boolean
            self._packed = None
            self._NaN = NaN # None
            self.extend(_read_file(self))
    
    def __setitem__(self, index, item):
        with self._lock:
            self._packed = None
            if isinstance(item, (list, set, tuple)):
                item = list(item)
            super(Maguro, self).__setitem__(index, item)
            
            autosave = isinstance(self._autosave, bool) and bool(self._autosave)
            if autosave:
                _write_file(self)
    
    def append(self, item, unique=False):
        with self._lock:
            self._packed = None
            if isinstance(item, (list, set, tuple)):
                item = list(item)
            if not unique or (item not in self):
                super(Maguro, self).append(item)

            autosave = isinstance(self._autosave, bool) and bool(self._autosave)
            if autosave:
                _write_file(self)
    
    def clear(self):
        with self._lock:
            self._packed = None
            super(Maguro, self).clear()

            autosave = isinstance(self._autosave, bool) and bool(self._autosave)
            if autosave:
                _write_file(self)
    
    def extend(self, iterable):
        with self._lock:
            self._packed = None
            super(Maguro, self).extend(iterable)

            autosave = isinstance(self._autosave, bool) and bool(self._autosave)
            if autosave:
                _write_file(self)
    
    def insert(self, index, item):
        with self._lock:
            self._packed = None
            super(Maguro, self).insert(index, item)

            autosave = isinstance(self._autosave, bool) and bool(self._autosave)
            if autosave:
                _write_file(self)
    
    def pop(self, pos):
        with self._lock:
            self._packed = None
            temp = super(Maguro, self).pop(pos)

            autosave = isinstance(self._autosave, bool) and bool(self._autosave)
            if autosave:
                _write_file(self)
            return temp
    
    def remove(self, item):
        with self._lock:
            self._packed = None
            super(Maguro, self).remove(item)

            autosave = isinstance(self._autosave, bool) and bool(self._autosave)
            if autosave:
                _write_file(self)
    
    def reverse(self):
        with self._lock:
            self._packed = None
            super(Maguro, self).reverse()

            autosave = isinstance(self._autosave, bool) and bool(self._autosave)
            if autosave:
                _write_file(self)
    
    def sort(self):
        with self._lock:
            self._packed = None
            super(Maguro, self).sort()

            autosave = isinstance(self._autosave, bool) and bool(self._autosave)
            if autosave:
                _write_file(self)
    
    def load(self, iterable):
        """ Replace old data with new contents. """
        with self._lock:
            self._packed = None
            if isinstance(iterable, (list, set, tuple)):
                try:
                    self.clear()
                    self.extend(list(iterable))

                    autosave = isinstance(self._autosave, bool) and bool(self._autosave)
                    if autosave:
                        _write_file(self)
                except:
                    pass
    
    def distinct(self):
        """ Remove duplicates on lists with hashable contents. """
        with self._lock:
            self._packed = None
            try:
                temp = list(set(self))
                self.clear()
                self.extend(temp)

                autosave = isinstance(self._autosave, bool) and bool(self._autosave)
                if autosave:
                    _write_file(self)
            except:
                pass
    
    def is_empty(self):
        """ Check if list object is empty or not. """
        with self._lock:
            return not len(self)

    def pack(self):
        """ Return delimiter-separates values in string representation. """
        with self._lock:
            self._packed = _pack(self)
            return self._packed
    
    def unpack(self):
        """ Return raw list object """
        with self._lock:
            return list(self)
    
    def save(self, save_as=None):
        """ Manual save to optimize operations. """
        with self._lock:
            if isinstance(save_as, str):
                self._filepath = save_as
            _write_file(self)

def _encode_dsv_equivalent(value, separator, quote_strings):
    """ Encode values into formatted string representations. """
    if value is None:
        return "null"
    if value == "NaN":
        return "NaN"
    if type(value) in (int, float):
        return value
    if isinstance(value, bool):
        if value:
            return "Yes"
        return "No"
    formatted = f"\"{value}\""
    if quote_strings:
        return formatted
    if separator is not None:
        if separator in value:
            return formatted
    if " " in value:
        return formatted
    return value

def _pack(data):
    """ Convert list object into delimiter-separated values representation. """
    if not isinstance(data._delimiter, str):
        return
    sep1 = data._delimiter
    sep2 = None
    if isinstance(data._newline, str):
        sep1 = data._newline
        sep2 = data._delimiter

    formatted = []
    quote_strings = isinstance(data._quote_strings, bool) and bool(data._quote_strings)
    for value in list(data):
        if sep2 is None:
            formatted.append(str(_encode_dsv_equivalent(value, sep1, quote_strings)))
            continue
        temp = []
        for part in value:
            temp.append(str(_encode_dsv_equivalent(part, sep2, quote_strings)))
        formatted.append(sep2.join(temp))
    return sep1.join(formatted)

def _clean_split(string, separator, allow_boolean, NaN):
    """ Smartly split valid strings outside quuted values. """
    formatted = string.replace(" ", "&nbsp;")
    formatted = formatted.replace(separator, "    &tmp;")
    parts = list(shlex.split(formatted))
    for i in range(len(parts)):
        temp = parts[i].replace("&nbsp;", " ").strip()
        if "    &tmp;" in temp:
            temp = temp.replace("    &tmp;", separator)
        else:
            temp = temp.replace("&tmp;", "")
        parts[i] = _autoparse_to_data_type(temp, allow_boolean, NaN)
    return parts

def _autoparse_to_data_type(string, allow_boolean, NaN):
    """ Decode DSV strings into equivalent Python data types. """
    formatted = string.strip()
    if len(formatted):
        if formatted == "null":
            return None
        if formatted == "NaN":
            return NaN
        if allow_boolean:
            if formatted.lower() in ("yes", "y"):
                return True
            if formatted.lower() in ("no", "n"):
                return False
        try:
            numeric = float(formatted)
            if not (numeric-(numeric//1)):
                return int(numeric)
            return numeric
        except:
            pass
    return formatted

def _write_file(data):
    """ Save formatted DSV into file. """
    if not (isinstance(data._filepath, str) and isinstance(data._encoding, str)):
        return
        
    packed = data._packed
    if not isinstance(packed, str):
        packed = _pack(data)

    try:
        with open(data._filepath, "w+", encoding=data._encoding) as f:
            f.write(packed)
    except:
        print("MaguroWarning: Unable to update selected file.")

def _read_file(data):
    """ Read DSV file and attempt to create a list object. """
    if not (isinstance(data._filepath, str) and isinstance(data._encoding, str)):
        return
    if not isinstance(data._delimiter, str):
        return
    sep1 = data._delimiter
    sep2 = None
    if isinstance(data._newline, str):
        sep1 = data._newline
        sep2 = data._delimiter

    temp = []
    allow_boolean = isinstance(data._allow_boolean, bool) and bool(data._allow_boolean)
    try:
        with open(data._filepath, "r", encoding=data._encoding) as f:
            values = f.read().split(sep1)
            for value in values:
                if len(value:=value.strip()):
                    if sep2 is not None:
                        temp.append(_clean_split(value, sep2, allow_boolean, data._NaN))
                        continue
                    temp.append(_autoparse_to_data_type(value, allow_boolean, data._NaN))
    except:
        pass
    return temp


from maguro import Maguro


print("\nTest #1a Initialization")
test = Maguro("temp/01a-initialization.csv")

print("\nTest #1b Auto Save = False")
test = Maguro("temp/01b-auto-save-no.csv", autosave=False)
for x in range(100):
    test.append(x, unique=True)
test.save() # save to initialized filepath
# save as new file
test.append(100, unique=True)
test.append(100)
test.append(100)
test.save(save_as="temp/01b-auto-save-new.csv")

print("\nTest #2a Native Set/Get")
test = Maguro("temp/02a-native-methods.csv")
test.append("Juan")
test[0] = "Pedro"
print(test[0])

print("\nTest #2b Native Clear")
test = Maguro("temp/02b-native-clear.csv")
test.append("Juan")
test.clear()
test.append("Pedro")
print(test.pack())

print("\nTest #2c Native Extend")
test = Maguro("temp/02c-native-extend.csv")
test.clear()
test.append("Juan")
test.extend(["Pedro", "Maria"])
print(test.pack())

print("\nTest #2d Native Insert")
test = Maguro("temp/02d-native-insert.csv")
test.clear()
test.append("Juan")
test.extend(["Pedro", "Maria"])
test.insert(1, "Soledad")
print(test.pack())

print("\nTest #2e Native Pop")
test = Maguro("temp/02e-native-pop.csv")
test.clear()
test.extend(["Juan", "Pedro", "Maria"])
temp = test.pop(2)
print(" - Pop:", temp)

print("\nTest #2f Native Remove")
test = Maguro("temp/02f-native-remove.csv")
test.clear()
test.extend(["Juan", "Pedro", "Maria"])
test.remove("Pedro")
print(test.pack())

print("\nTest #2g Native Reverse")
test = Maguro("temp/02g-native-remove.csv")
test.clear()
test.extend(["Juan", "Pedro", "Maria"])
test.reverse()
print(test.pack())

print("\nTest #2h Native Sort")
test = Maguro("temp/02g-native-sort.csv")
test.clear()
test.extend(["Juan", "Pedro", "Maria"])
test.sort()
print(test.pack())

print("\nTest #3a 1D array")
test = Maguro("temp/03a-1d.csv", delimiter=",")
print(test.unpack())

print("\nTest #3b 2D array")
test = Maguro("temp/03b-2d.csv", delimiter=",", newline="\n")
print(test.unpack())

print("\nTest #3c Mixed-data 2D array")
test = Maguro("temp/03c-2d-mixed.csv", delimiter=",", newline="\n")
print(test.unpack())

print("\nTest #4 Booleans")
test = Maguro("temp/04-booleans.csv", delimiter=",", newline="\n", allow_boolean=True)
print(test.unpack())

print("\nTest #5a NaNs")
test = Maguro("temp/05a-NaNs.csv", delimiter=",", newline="\n")
print(test.unpack())

print("\nTest #5b NaNs as ''")
test = Maguro("temp/05b-NaNs.csv", delimiter=",", newline="\n", NaN="")
print(test.unpack())

print("\nTest #5c NaNs as 0")
test = Maguro("temp/05c-NaNs.csv", delimiter=",", newline="\n", NaN=0)
print(test.unpack())

print("\nTest #6 Format to DSV")
test = Maguro("temp/06-format-to-dsv.csv", delimiter=",", newline="\n")
print(test.pack())

print("\nTest #7 Remove Duplicates")
test = Maguro("temp/07-distinct.csv")
test.clear()
test.extend(["Juan", "Pedro", "Maria", 1, "2", 2])
test.append("Juan")
test.distinct()
print(test.unpack())

print("\nTest #8 Check if Empty or Not")
test = Maguro("temp/8-is-empty.csv")
test.clear()
print("Empty: ", test.is_empty())
test.extend(["Juan", "Pedro", "Maria", 1, "2", 2])
print("Empty: ", test.is_empty())

print("\nTest #9 TSV file")
test = Maguro("temp/9-tab-separated-values.tsv", delimiter="\t", newline="\n", quote_strings=True)
test.clear()
test.append(["a", "b", "c", "d", "e"])
test.append(["1", "2", "3", "4", "5"])
test.append([1, 2, 3, 4, 5])
print(test.unpack())

print("\nTest #10 Load and replace new contents.")
test = Maguro("temp/10-load-list.tsv", delimiter="\t", newline="\n", quote_strings=True)
test.clear()
test.append(["a", "b", "c", "d", "e"])
print("Test:", test.unpack())
test.load([["1", "2", "3", "4", "5"]])
print("Test:",test.unpack())